const mimeTypeParser = require('content-type');

/**
 * JSON schema for additional parameters which aren't described as swagger parameters
 */
const additionalParameters = {
	'Content-Type': {
		title: 'Content-Type',
		description: 'This is the expected payload format that you send.',
		type: 'string',
		group: 'Advanced HTTP Options'
	},
	Accept: {
		title: 'Accept',
		description: 'This is the response format you expect to receive.',
		type: 'string',
		group: 'Advanced HTTP Options'
	},
	'Decode Response': {
		title: 'Decode Response',
		description: 'Controls how the response is decoded. If auto, a best-effort attempt will be made to decode the response according to its content-type header. If the content-type is "application/json", the response will decoded as JSON but may fail if the data is unparsable. If "text/*" or "application/xml" it will be returned as a string. For any other content-types and unknown charsets, the response be will be returned as a NodeJS Buffer. If binary, no decoding is attempted and the response will be returned as a NodeJS Buffer.',
		type: 'string',
		enum: [ 'auto', 'binary' ],
		default: 'auto',
		group: 'Advanced HTTP Options'
	}
};

const plusJSONregex = /\+?json$/;

/**
 * @typedef {Object} Parameter
 * @property {string} id the id of the parameter
 * @property {Object} value The json schema for the parameter
 */

/**
 * Constructs and returns additional parameters that will be rendered as part of a method.
 * Some of them are based on Swagger `consumes` and `produces` per method or on global level.
 * Others are pure configuration properties.
 *
 * @param {object} method the method from the swagger.
 * @param {object} payloadInfo information about the documented payload of the request
 * @param {string} payloadInfo.type body, formData or undefined depending on how the payload is described.
 * @param {boolean} payloadInfo.hasFiles true if the method has file parameters
 * @returns {Parameter[]} array of additional parameters
 */
function getAdditionalParameters(method, payloadInfo) {
	const parameters = [];

	// Calculate content type parameter based on the swagger
	const contentType = getContentType(payloadInfo, method.consumes);
	if (contentType) {
		parameters.push(contentType);
	}

	// Calculate accept parameter based on the swagger
	const accept = getAccept(method.produces);
	parameters.push(accept);

	// Always add decode response parameter
	parameters.push({
		id: '_decode-response',
		value: additionalParameters['Decode Response']
	});
	return parameters;
}

/**
 * Calculate and return schema for the content type parameter
 * @param {object} payloadInfo information about the documented payload of the request
 * @param {string} payloadInfo.type body, formData or undefined depending on how the payload is described.
 * @param {boolean} payloadInfo.hasFiles true if the method has file parameters
 * @param {string[]} consumes array of consumes headers documented for current method
 * @returns {Parameter} the content type parameter
 *                   undefined if the parameter should not be included
 */
function getContentType(payloadInfo, consumes = []) {
	// Content type relevant only if we have formData or body payload
	if (!payloadInfo.type) {
		return;
	}

	const parsedMimeTypes = parseMimeTypes(consumes);
	const validConsumes = parsedMimeTypes.map(t => t.original);

	const parameter = {
		...additionalParameters['Content-Type']
	};

	// If the swagger provides a number of consumes, allow them as options
	if (validConsumes.length) {
		parameter.enum = validConsumes;
	}

	if (payloadInfo.type === 'formData') {
		const multipart = 'multipart/form-data';
		const urlencoded = 'application/x-www-form-urlencoded';

		if (validConsumes.length) {
			// There are provided options so we shouldn't change them. but pick a preferred one
			const preference = payloadInfo.hasFiles ? multipart : urlencoded;
			parameter.default = getFormDataMimeType(parsedMimeTypes, preference) || validConsumes[0];
		} else if (payloadInfo.hasFiles) {
			// If we have files in the payload then add multipart
			parameter.default = multipart;
			parameter.enum = [ multipart ];
		} else {
			// If we don't have files then add both content types and default to urlencoded
			parameter.default = urlencoded;
			parameter.enum = [ urlencoded, multipart ];
		}
	} else {
		// other body payloads default to the closest JSON type in consumes
		// if none could be found, it uses the first provided type in consumes
		// otherwise it falls back to application/json
		parameter.default = getJSONMimeType(parsedMimeTypes) || validConsumes[0] || 'application/json';
	}

	return {
		id: '_content-type',
		value: parameter
	};
}

/**
 * Calculate and return schema for the accept parameter
 * @param {string[]} produces array of produces headers documented for current method
 * @returns {Parameter} the accept parameter
 *                   undefined if the parameter should not be included
 */
function getAccept(produces = []) {
	const parsedMimeTypes = parseMimeTypes(produces);
	const validProduces = parsedMimeTypes.map(t => t.original);

	const parameter = {
		...additionalParameters['Accept']
	};

	// If the swagger provides a number of values, allow them as options
	if (validProduces.length) {
		parameter.enum = validProduces;
	}

	// defaults to the closest JSON type in produces
	// if none could be found, it uses the first provided type in produces
	// otherwise it falls back to application/json
	parameter.default = getJSONMimeType(parsedMimeTypes) || validProduces[0] || 'application/json';

	return {
		id: '_accept',
		value: parameter
	};
}

function parseMimeTypes(mimeTypes) {
	return mimeTypes
		.map(parseMimeType)
		.filter(t => t !== undefined);
}

/**
 * Gets JSON mime type from array of mime type metadata
 * Order of preference: application/json, XXX/XXX+json
 *
 * @param {object[]} parsedMimeTypes array of parsed mime types.
 * @returns {string} the JSON mime type found. If none are found, returns undefined.
 */
function getJSONMimeType(parsedMimeTypes) {
	let type;
	for (const parsedType of parsedMimeTypes) {
		const json = parsedType.type === 'application/json';
		const plusJSON = parsedType.type.match(plusJSONregex) !== null;
		if (json || plusJSON) {
			type = parsedType.original;
		}
		if (json) {
			break;
		}
	}
	return type;
}

/**
 * Gets a form data mime type from list of possible mime types
 * will prioritise a preference if exists
 *
 * @param {object[]} parsedMimeTypes array of parsed mime types.
 * @param {string} preference preferred mime type
 * @returns {string} the form-data mime type found. If none are found, returns undefined.
 */
function getFormDataMimeType(parsedMimeTypes, preference) {
	let type;
	for (const parsedType of parsedMimeTypes) {
		const urlencoded = parsedType.type === 'application/x-www-form-urlencoded';
		const multipart = parsedType.type === 'multipart/form-data';
		if (parsedType.type === preference) {
			return parsedType.original;
		} else if (urlencoded || multipart) {
			type = parsedType.original;
		}
	}
	return type;
}

/**
 * Parse the mime type with the help of content-type library.
 *
 * @param {string} mimeType the mime type
 * @returns {object} object with the information for the parsed mime type or undefined in case of exception
 */
function parseMimeType(mimeType) {
	try {
		const parsedType = mimeTypeParser.parse(mimeType);
		parsedType.original = mimeType;
		return parsedType;
	} catch (e) {
		// content-type library could fail on wrong types so returned undefined in that case
		// https://github.com/jshttp/content-type#contenttypeparsestring
		return undefined;
	}
}

module.exports = {
	getContentType,
	getAccept,
	getAdditionalParameters
};
